-- Co-op 2.2
-- By Jon Irons
-- jonirons@gmail.com

-- user variables

-- beacon as it appears physically
BEACON = "alien short light"

-- beacon placement sound
BEACON_SOUND = "computer login"

-- "teleporting" out leaves this behind:
OUT_EFFECT = "major defender detonation"

-- same effect for teleport in by default
IN_EFFECT = OUT_EFFECT

-- sounds for teleport in and out
OUT_SOUND = "teleport out"
IN_SOUND = "teleport in"

-- this is the fader effect when you teleport
FADER = "negative"

-- modify this list if your scenario needs to change items
-- that shouldn't be shared among players for gameplay
-- balance reasons.
NON_SHARE_ITEMS = {
   "invisibility",
   "invincibility",
   "infravision",
   "key",
   "uplink chip"
}

-- DO NOT MODIFY BELOW THIS LINE. ESPECIALLY YOU, TIM.

-- hacker sections: "constants," "main," and "icons"

CollectionsUsed = {4,26}
Triggers = {}

-- forward compatibility
if not math.mod then
   math.mod = function (a, b) return a % b end
end

-- constants
INVINC = 2 * 30
MAX_OVERLAY_STRING = 12
INFO_TIME = 20
BEACON_DELAY = 60
MP = "%-%-[Mm]" -- monster info pattern
AGE_X = 0.5
AGE_Y = 0.5 -- y offset of beacon age
SOLO_DEBUG = true
MONSTERS = 0

-- main
function Triggers.init()
   if Game.type ~= "cooperative play" and
      not SOLO_DEBUG then return end
   for p in Players() do
      p._beacon = {}
      p._beacon.owner = p
      set_player_beacon(p, p.monster.x,
			p.monster.y, p.monster.z,
			p.monster.polygon, p.direction,
			p.elevation)
      p._overhead_active = false
      p._beacon.age = 0
      p._beacon.delay = 0
      p._info_time = 0
      p._last_target = ""
      p._beackpeack = {}
      p._share = true
      p._receive = true
      p._kills = 0
      --prevent save game restore from sapping inventory?
      --if not p.dead then fill_beackpeack(p) end
      fill_beackpeack(p)
      p._monster_info = true
      if string.match(p.name, MP) then
	 p._monster_info = false
      end
   end
end

function Triggers.idle()
   if Game.type ~= "cooperative play" and
      not SOLO_DEBUG
   then
      if not already then
	 Players.print("Game needs to be hosted as Co-Op.")
	 already = true
      end
      return
   end

   MONSTERS = 0
   for m in Monsters() do
      if m.valid and not
	 m.type.friends[Players[0].monster.type.class.mnemonic]
      then
	 MONSTERS = MONSTERS + 1
      end
   end

   for p in Players() do
      p._beacon.age = p._beacon.age + 1
      detect_overhead_map(p)
      detect_overhead_catchup(p)
      detect_beacon_set(p)
      detect_beacon_cycle(p)
      hud(p)
      beacon_info(p)
   end
end

function Triggers.player_killed(player, killer, action, projectile)
   fill_beackpeack(player)
   empty_inventory(player)
   player._overhead_active = false
   player.weapons.active = false
end

function Triggers.player_revived(player)
   empty_beackpeack(player)
   if Level.rebellion then player.life = 150 end
   player.weapons.active = true
end

function Triggers.got_item(item, player)
   local can_share = true
   for k, v in pairs(NON_SHARE_ITEMS) do
      if item.mnemonic == v then can_share = false end
   end

   if can_share then
      for p in Players() do
	 if p ~= player and
	    player._share and
	    p._receive then
	    p._share = false
	    p.items[item] = p.items[item] + 1
	    p._share = true
	 end
      end
   end
end

function Triggers.monster_killed(monster, killer, projectile)
   if killer and
      (not monster.type.friends["player"] or
       not killer.monster.type.friends[monster.type.class])
   then
      killer._kills = killer._kills + 1
   end
end

function Triggers.cleanup()
   for p in Players() do
      if p.dead then empty_beackpeack(p) end
   end
end

function detect_overhead_map(player)
   if player.action_flags.toggle_map and not player.dead then
      player._overhead_active = not player._overhead_active
   end
end

function detect_overhead_catchup(player)
   if player._beacon.delay > 0 then
      player._beacon.delay = player._beacon.delay - 1
   end
   if player._overhead_active and
      player._beacon.delay == 0 and
      player.action_flags.microphone_button then
      player.action_flags.microphone_button = false
      send_to_beacon(player, player._beacon.owner)
      player._beacon.delay = BEACON_DELAY
   end

   if player.action_flags.microphone_button and
      player._overhead_active then
      player.action_flags.microphone_button = false
   end
end

function detect_beacon_set(player)
   if player.action_flags.action_trigger and
      player._overhead_active and
      not player.dead and
      not player:find_action_key_target() then
      set_player_beacon(player, player.monster.x,
			player.monster.y, player.monster.z,
			player.monster.polygon,
			player.monster.yaw, 0)
   end
end   

function set_player_beacon(player, x, y, z, p, dir, ele)
   player._beacon.x = x
   player._beacon.y = y
   player._beacon.z = z
   player._beacon.p = p
   player._beacon.dir = dir
   player._beacon.ele = ele
   player._beacon.age = 0
   if player._beacon.object then
      player._beacon.object:delete() end
   player._beacon.object =
      Scenery.new(x, y,
		  (z - p.floor.z) + player.monster.type.height/2,
		  p, BEACON)
   player._beacon.object.solid = false
   for guy in Players() do
      guy:play_sound(BEACON_SOUND, 0.75)
      if guy ~= player then
	 guy:print("-> "..strip_name(player).." set a beacon.")
      end
   end
   player._icon_flash = FLASH_TIME
end

function send_to_beacon(player, owner)
   player.monster:play_sound(OUT_SOUND)
   Effects.new(player.monster.x,
	       player.monster.y,
	       player.monster.z + 0.45,
	       player.monster.polygon,
	       OUT_EFFECT)
   player.invincibility_duration = INVINC
   player:fade_screen(FADER)
   player:position(owner._beacon.x,
		   owner._beacon.y,
		   owner._beacon.z,
		   owner._beacon.p)
   player.monster:play_sound(IN_SOUND)
   player.direction = player._beacon.dir
   player.elevation = player._beacon.ele
end

function detect_beacon_cycle(player)
   if player._overhead_active or player.dead then
      if player.action_flags.cycle_weapons_forward then
	 player.action_flags.cycle_weapons_forward = false
	 local i = player._beacon.owner.index + 1
	 if i > #Players - 1 then i = 0 end
	 player._beacon.owner = Players[i]
	 player.monster:play_sound(BEACON_SOUND)
      elseif player.action_flags.cycle_weapons_backward then
	 player.action_flags.cycle_weapons_backward = false
	 local i = player._beacon.owner.index - 1
	 if i < 0 then i = #Players - 1 end
	 player._beacon.owner = Players[i]
	 player.monster:play_sound(BEACON_SOUND)
      end
   end
end   

function fill_beackpeack(player)
   for items in ItemTypes() do
      local item = items.mnemonic
      player._beackpeack[item] = player.items[item];   
   end
end

function empty_inventory(player)
   for items in ItemTypes() do
      if items ~= "fist" and
	 items ~= "knife" and
	 items ~= "pistol" then
	 player._share, player._receive = false, false
	 player.items[items.mnemonic] = 0
	 player._share, player._receive = true, true
      end
   end
end

function empty_beackpeack(player)
   for k,v in pairs(player._beackpeack) do
      local item = k
      player._share, player._receive = false, false
      player.items[k] = v
      player._share, player._receive = true, true
      player._beackpeack[item] = 0
   end
end

function hud(player)
   overlay0(player)
   overlay1(player)
   overlay2(player)
   overlay3(player)
   overlay4(player)
   overlay5(player)
end

function overlay0(player)
   local text = " "
   local icon = nil
   if player._overhead_active or player.dead then
      if player._beacon.owner ~= player then
	 text = truncate(strip_name(player._beacon.owner))
	 icon = ICONS.slave
      else
	 icon = ICONS.beacon
      end
   else
      local target, o = player:find_target()
      if is_monster(target) then
	 local life = target.life
	 if life < 0 then life = "" end
	 icon = ICONS.eye
	 if target.player then
	    text = strip_name(target.player).." "..target.player.life
	    icon = string.gsub(icon, eye_color,
				 hex[target.player.color.index], 1)
	 elseif player._monster_info then
	    player._last_target =
	       "["..tostring(target.type).."] "..life
	    text = player._last_target
	    player._info_time = INFO_TIME
	 end
      elseif player._info_time > 0 then
	 player._info_time = player._info_time - 1
	 icon = ICONS.eye
	 text = player._last_target
      end
   end
   player.overlays[0].text = text
   player.overlays[0].icon = icon
end

function overlay1(player)
   local text = " "
   local icon = nil
   if player._overhead_active then
      icon = ICONS.prev
      text = "SLAVE--"
   else
      if player.dead then
	 icon = ICONS.prev
	 text = "SLAVE--"
      end
   end
   player.overlays[1].text = text
   player.overlays[1].icon = icon
end

function overlay2(player)
   local text = " "
   local icon = nil
   if player._overhead_active then
      icon = ICONS.next
      text = "SLAVE++"
   else
      if player.dead then
	 icon = ICONS.next
	 text = "SLAVE++"
      end
   end
   player.overlays[2].text = text
   player.overlays[2].icon = icon
end

function overlay3(player)
   local text = " "
   local icon = nil
   if player._overhead_active then
      icon = ICONS.action
      text = "SET BEACON"
   else
   end
   player.overlays[3].text = text
   player.overlays[3].icon = icon
end

function overlay4(player)
   local text = " "
   local icon = nil
   if player._overhead_active then
      text = "TELEPORT"
      icon = ICONS.mic
   else
      text = player._kills.." : "..MONSTERS
      icon = ICONS.monsters
   end
   player.overlays[4].text = text
   player.overlays[4].icon = icon
end

function overlay5(player)
   local text = " "
   local icon = nil
   if player._overhead_active then
   else
   end
   player.overlays[5].text = text
   player.overlays[5].icon = icon
end

function fmt_time_minutes(time)
   return pad(math.floor(time/(30*60)))..":"..
   pad(math.mod(math.floor(time/30), 60))
end

function pad(n)
   if n < 10 then return "0"..n end
   return n
end

function truncate(str)
   if string.len(str) > MAX_OVERLAY_STRING then
      return string.sub(str, 1,
			MAX_OVERLAY_STRING-3).."..."
   end
   return str
end

function beacon_coords(player)
   local b = player._beacon
   return (math.floor(b.x * 10)/10).." , "..
   (math.floor(b.y * 10)/10)
end

function strip_name(player)
   return string.gsub(player.name, MP, "")
end

function beacon_info(player)
   local str = ". "..player.name
   local beacon = player._beacon
   local age = fmt_time_minutes(beacon.age)
   local annotations = beacon.annotations
   if not annotations or not annotations.position or
      not annotations.age then
      annotations = {}
      annotations.position =
	 Annotations.new(beacon.p, str,
			 beacon.x, beacon.y)
      annotations.age =
	 Annotations.new(beacon.p, age,
			 beacon.x + AGE_X,
			 beacon.y + AGE_Y)
      beacon.annotations = annotations
   else
      local a_pos = annotations.position
      local a_age = annotations.age
      a_pos.x, a_pos.y, a_pos.polygon =
	 beacon.x, beacon.y, beacon.p
      a_age.x, a_age.y, a_age.polygon =
	 beacon.x + AGE_X, beacon.y + AGE_Y,
      beacon.p
   end
   annotations.position.text = str
   annotations.age.text = age
end

-- icons

ICONS = {}

ICONS.slave = [[
2
.00000000
 ffff00ff
................1
......     .....2
.....   .   ....3
....   ....  ...4
....   ..... ...5
....   .........6
.....   ........7
......    ......8
.......    .....9
.........   ....0
..........   ...1
.... .....   ...2
....  ....   ...3
.....   .   ....4
......     .....5
................6
1234567890123456
]]

ICONS.beacon =[[
2
.00000000
 00ff00ff
................1
....      ......2
...   .    .....3
...   ..    ....4
...   ...   ....5
...   ...   ....6
...   ..   .....7
...       ......8
...   .   ......9
...   ..   .....0
...   ...   ....1
...   ...   ....2
...   ..    ....3
...        .....4
....     .......5
................6
1234567890123456
]]

ICONS.eye = [[
5
.00000000
 ffffffff
#ff00bbff
O000000ff
*aa0000ff
.....      .....1
...  ** *    ...2
..     *      ..3
.    #####     .4
.   #######    .5
 * ###OOO###    6
 * ##OOOOO##* * 7
  *##OOOOO## *  8
   ##OOOOO##  * 9
   ###OOO###  * 0
   *#######  *  1
.  * #####     .2
. *      *     .3
.. *    *     ..4
...    * *   ...5
.....      .....6
1234567890123456
]]

eye_color = 'ff00bb'

hex = {}
hex.slate, hex[0] = '205da0', '205da0'
hex.red, hex[1] = 'f80000', 'f80000'
hex.violet, hex[2] = 'b00258', 'b00258'
hex.yellow, hex[3] = 'f9f800', 'f9f800'
hex.white, hex[4] = 'e9efe8', 'e9efe8'
hex.orange, hex[5] = 'f05800', 'f05800'
hex.blue, hex[6] = '0807f8', '0807f8'
hex.green, hex[7] = '01f800', '01f800'

ICONS.mic = [[
10
 00000000
.eeeeeeff
*ddddddff
occccccff
Obbbbbbff
@aaaaaaff
+999999ff
$888888ff
#777777ff
&bb11ddff
       #     &  
      $$   &  & 
     +++    &  &
    @@@@  &  & &
OOOOOOOO   & & &
oooooooo & & & &
******** & & & &
........ & & & &
........ & & & &
******** & & & &
oooooooo & & & &
OOOOOOOO   & & &
    @@@@  &  & &
     +++    &  &
      $$   &  & 
       #     &  
]]
ICONS.monsters = [[
35
 000000
.020F08
+270000
@14090B
#3B0000
$001B1F
%14140F
&031D10
*4E0000
=620000
-052C18
;760000
>29271E
,063B1F
a890000
)432322
!2D2D2D
~9D0000
{084927
]B10000
^3D3B2D
/0A582F
(D80000
_0B6737
:514E3C
<137C45
[66624B
}1A9153
|7A765A
122A661
28E8969
3A29D78
4B7B087
5E1DDC3
600000000
666666%&>&%66666
66666$[332[&6666
66666|42][3[6666
666623^;5=^2:666
66623[>a~]:||^66
66^[:];#;+aa[^>6
66>^=5(+)=5]+^>6
66&%#==*!#*a+>%6
66%>@@+>{{##%%&6
66%>>^{_<}1}-%%6
66%%-<_&<&}1&%%6
666&& _-_&}<.666
666&&._-/&<{.666
66.&&./,{&<{&666
.-.&..&{/.<{-&.<
<&.&..&-___{-&..
]]

--the following icons are from Visual Mode Lua
ICONS.next =
[[
8
 000000
.DD0000
+777777
@FF6633
#00BB00
$888888
%AAAAAA
&FFFFFF00
&&&&&&&&@..@&&&&
&&&&&&&@.$$.@&&&
&&&&&&@.$$&%.$&&
&&&&&&.$$&&&.$&&
&&&&&&.$&&.....&
&&&&&&.$&&&...$$
&&&&&&.$&&&&.$$&
&&&&&&&$&&&&&$&&
&&&&&&&&&&&&&&&&
   &&    &&    &
 # +& ## +& ## +
 # +& ## +& ## +
 # +& ## +& ## +
 # +& ## +& ## +
   +&    +&    +
&+++&&++++&&++++
]]

ICONS.prev =
[[
8
 000000
.DD0000
+777777
@FF6633
#00BB00
$888888
%AAAAAA
&FFFFFF00
&&&@..@&&&&&&&&&
&&@.$$.$&&&&&&&&
&&.$$&%.$&&&&&&&
&&.$&&&%.%&&&&&&
.....&&&.$&&&&&&
&...$$&&.$&&&&&&
&&.$$&&&.$&&&&&&
&&&$&&&&&%&&&&&&
&&&&&&&&&&&&&&&&
    &&    &&   &
 ## +& ## +& # +
 ## +& ## +& # +
 ## +& ## +& # +
 ## +& ## +& # +
    +&    +&   +
&++++&&++++&&+++
]]

ICONS.action =
[[
3
 000000
.00FFFF00
+FFFFFF
.......  .......
...  . ++   ....
.. ++  ++ ++ ...
.. ++  ++ ++ . .
... ++ ++ ++  + 
... ++ ++ ++ ++ 
.  . +++++++ ++ 
 ++  ++++++++++ 
 +++ +++++++++ .
. ++++++++++++ .
.. +++++++++++ .
.. ++++++++++ ..
... +++++++++ ..
.... +++++++ ...
..... ++++++ ...
..... ++++++ ...
]]
